///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2004 Intel Corporation 
// All rights reserved. 
//
// Redistribution and use in source and binary forms, with or without 
// modification, are permitted provided that the following conditions are met: 
//
// * Redistributions of source code must retain the above copyright notice, 
// this list of conditions and the following disclaimer. 
// * Redistributions in binary form must reproduce the above copyright notice, 
// this list of conditions and the following disclaimer in the documentation 
// and/or other materials provided with the distribution. 
// * Neither name of Intel Corporation nor the names of its contributors 
// may be used to endorse or promote products derived from this software 
// without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR 
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ipp.h"

#include "filter.h"
#include "Image.h"

Filter::Filter() : isSrcROI_(0), isDstROI_(0), borderMode_(0),
	borderVal_(0)
{
}

Filter::~Filter() {}

void Filter::SetBorderMode(int type) { borderMode_ = type; }

int Filter::GetBorderMode() const { return borderMode_; }

void Filter::SetSrcROI(IppiRect srcROI)
{
	srcROI_.x = srcROI.x;
	srcROI_.y = srcROI.y;
	srcROI_.width = srcROI.width;
	srcROI_.height = srcROI.height;

	isSrcROI_ = 1;
}

//add destWindow? (differentiate between ROI and shift)
//maybe just "set shift" that overrides default shift
void Filter::SetDestROI(IppiRect dstROI)
{
	dstROI_.x = dstROI.x;
	dstROI_.y = dstROI.y;
	dstROI_.width = dstROI.width;
	dstROI_.height = dstROI.height;
	
	isDstROI_ = 1;
}

IppiRect Filter::GetSrcRect(IppiRect dstRect) const
{
	int xMin, yMin, xMax, yMax;

	xMin = dstRect.x - GetXMinBorder();
	yMin = dstRect.y - GetYMinBorder();
	xMax = dstRect.x + dstRect.width + GetXMaxBorder();
	yMax = dstRect.y + dstRect.height + GetYMaxBorder();

	if (isDstROI_)
	{
		xMin = IPP_MAX(xMin, dstROI_.x);
		yMin = IPP_MAX(yMin, dstROI_.y);
		xMax = IPP_MIN(xMax, dstROI_.x + dstROI_.width);
		yMax = IPP_MIN(yMax, dstROI_.y + dstROI_.height);
	}

	if (isSrcROI_)
	{
		xMin = IPP_MAX(xMin, srcROI_.x);
		yMin = IPP_MAX(xMin, srcROI_.y);
		xMax = IPP_MIN(xMax, srcROI_.x + srcROI_.width);
		yMax = IPP_MIN(yMax, srcROI_.y + srcROI_.height);
	}

	IppiRect srcRect;
	srcRect.x = xMin;
	srcRect.y = yMin;
	srcRect.width = xMax - xMin;
	srcRect.height = yMax - yMin;

	return srcRect;
}

IppiSize Filter::GetSrcSize(IppiSize dstSize) const
{
	IppiRect dstRect = { 0,0,dstSize.width, dstSize.height };
	IppiRect srcRect = GetSrcRect(dstRect);
	IppiSize srcSize = { srcRect.width, srcRect.height };
	return srcSize;
}

IppiSize Filter::GetSrcSize() const
{
	IppiSize srcSize = {0,0};

	if (isDstROI_)
	{
		IppiRect srcRect = GetSrcRect(dstROI_);
		srcSize.width = srcRect.width;
		srcSize.height = srcRect.height;
	}
	return srcSize;
}

IppiRect Filter::GetDestRect(IppiRect srcRect) const
{
	int xMin, yMin, xMax, yMax;

	xMin = srcRect.x + GetXMinBorder();
	yMin = srcRect.y + GetYMinBorder();
	xMax = srcRect.x+srcRect.width - GetXMaxBorder();
	yMax = srcRect.y+srcRect.height - GetYMaxBorder();

	if (isSrcROI_)
	{
		xMin = IPP_MAX(xMin, srcROI_.x);
		yMin = IPP_MAX(yMin, srcROI_.y);
		xMax = IPP_MIN(xMax, srcROI_.x+srcROI_.width);
		yMax = IPP_MIN(yMax, srcROI_.y+srcROI_.height);
	}

	if (isDstROI_)
	{
		xMin = IPP_MAX(xMin, dstROI_.x);
		yMin = IPP_MAX(xMin, dstROI_.y);
		xMax = IPP_MIN(xMax, dstROI_.x + dstROI_.width);
		yMax = IPP_MIN(yMax, dstROI_.y + dstROI_.height);
	}

	IppiRect dstRect;
	dstRect.x = xMin;
	dstRect.y = yMin;
	dstRect.width = xMax - xMin;
	dstRect.height = yMax - yMin;

	return dstRect;
}

IppiSize Filter::GetDestSize(IppiSize srcSize) const
{
	IppiRect srcRect = {0,0,srcSize.width, srcSize.height};
	IppiRect dstRect = GetDestRect(srcRect);
	IppiSize dstSize = { dstRect.width, dstRect.height };
	return dstSize;
}

IppiSize Filter::GetDestSize() const
{
	IppiSize dstSize = {0,0};

	if (isSrcROI_)
	{
		IppiRect dstRect = GetDestRect(srcROI_);
		dstSize.width = dstRect.width;
		dstSize.height = dstRect.height;
	}
	return dstSize;
}

/*
int Filter::GetXMinBorder() const
   { return GetKernelSize().width - GetAnchor().x - 1; }
int Filter::GetXMaxBorder() const { return GetAnchor().x; }
int Filter::GetYMinBorder() const
   { return GetKernelSize().height - GetAnchor().y - 1; }
int Filter::GetYMaxBorder() const { return GetAnchor().y; }
*/
int Filter::GetXMinBorder() const
	{ return GetAnchor().x; }
int Filter::GetXMaxBorder() const
	{ return GetKernelSize().width - GetAnchor().x - 1; }
int Filter::GetYMinBorder() const
	{ return GetAnchor().y; }
int Filter::GetYMaxBorder() const
	{ return GetKernelSize().height - GetAnchor().y - 1; }

IppiRect Filter::GetFullSrcRect(IppiRect srcRect) const
{
	int xMin, yMin, xMax, yMax;

	xMin = srcRect.x - GetXMinBorder();
	yMin = srcRect.y - GetYMinBorder();
	xMax = srcRect.x + srcRect.width + GetXMaxBorder();
	yMax = srcRect.y + srcRect.height + GetYMaxBorder();

	if (isSrcROI_)
	{
		xMin = IPP_MAX(xMin, srcROI_.x);
		yMin = IPP_MAX(yMin, srcROI_.y);
		xMax = IPP_MIN(xMax, srcROI_.x+srcROI_.width);
		yMax = IPP_MIN(yMax, srcROI_.y+srcROI_.height);
	}

	IppiRect fullSrcRect;
	fullSrcRect.x = xMin;
	fullSrcRect.y = yMin;
	fullSrcRect.width = xMax - xMin;
	fullSrcRect.height = yMax - yMin;

	return fullSrcRect;
}

int Filter::MakeBorderImage(const Image8u* pSrc, Image8u* pBorderIm,
							int borderMode) const
{
	IppiSize srcSize = pSrc->GetSize();
	IppiRect srcRect = {0, 0, srcSize.width, srcSize.height};
	if (isSrcROI_)
	{
		srcRect = srcROI_;
		srcSize.width = srcROI_.width;
		srcSize.height = srcROI_.height;
	}

	IppiRect borderRect = GetFullSrcRect(srcRect);
	int nChannels = pSrc->GetChannelCount();
	if ((nChannels != 1) && (nChannels != 3) &&(nChannels != 4))
		return -1;

	IppiSize tmp = pBorderIm->GetSize();
	if ((tmp.width != borderRect.width) || (tmp.height != borderRect.height) ||
		(nChannels != pBorderIm->GetChannelCount()))
		pBorderIm->InitAlloc(borderRect.width, borderRect.height, nChannels);

	IppStatus st;
	Ipp8u* pBData = pBorderIm->GetData();
	const Ipp8u* pSData = pSrc->GetConstData();
	int bStride = pBorderIm->GetStride(), sStride=pSrc->GetStride();
	IppiSize bSize = pBorderIm->GetSize(), sSize = pSrc->GetSize();
	int xmin = GetXMinBorder(), xmax = GetXMaxBorder();
	int ymin = GetYMinBorder(), ymax = GetYMaxBorder();
	if (borderMode == VAL)
	{
		Ipp8u border3[3] = {borderVal_, borderVal_, borderVal_};
		Ipp8u border4[4] = {borderVal_, borderVal_, borderVal_, borderVal_};

		if (nChannels == 1) st = ippiCopyConstBorder_8u_C1R(
						pSData, sStride, sSize,
						pBData, bStride, bSize,
						ymin, xmin, borderVal_);
		else if (nChannels == 3) st = ippiCopyConstBorder_8u_C3R(
						pSData, sStride, sSize,
						pBData, bStride, bSize,
						ymin, xmin, border3);
		else if (nChannels == 4) st = ippiCopyConstBorder_8u_C4R(
						pSData, sStride, sSize,
						pBData, bStride, bSize,
						ymin, xmin, border4);
	}
	else if (borderMode == REPEAT)
	{
		if (nChannels == 1) ippiCopyReplicateBorder_8u_C1R(
						pSData, sStride, sSize,
						pBData, bStride, bSize,
						ymin, xmin);
		else if (nChannels == 3) ippiCopyReplicateBorder_8u_C3R(
						pSData, sStride, sSize,
						pBData, bStride, bSize,
						ymin, xmin);
		else if (nChannels == 4) ippiCopyReplicateBorder_8u_C4R(
						pSData, sStride, sSize,
						pBData, bStride, bSize,
						ymin, xmin);
	}
	else
	{
		int i;
		int w = pSrc->GetSize().width, h = pSrc->GetSize().height;

		pBorderIm->Zero();
		pBorderIm->CopyFrom_R( pSrc->GetConstData(), w,h,
							   pSrc->GetStride(),xmin,ymin);

		if (borderMode == REFLECT)
		{
			for (i=0; i<ymin; i++)
				pBorderIm->CopyFrom_R( pSrc->GetConstData(0, i+1),
					w,1, pSrc->GetStride(), xmin,ymin-i-1);
			for (i=0; i<ymax; i++)
				pBorderIm->CopyFrom_R( pSrc->GetConstData(0, h-1 - i-1),
					w,1, pSrc->GetStride(), xmin,h+ymin-1 + i+1);
			for (i=0; i<xmin; i++)
				pBorderIm->CopyFrom_R( pSrc->GetConstData(i+1,0),
					1,h, pSrc->GetStride(), xmin-i-1,ymin);
			for (i=0; i<xmax; i++)
				pBorderIm->CopyFrom_R( pSrc->GetConstData(w-1 - i-1,0),
					1,h, pSrc->GetStride(), w+xmin-1 + i+1,ymin);
		}
		else if (borderMode == WRAP)
		{
			pBorderIm->CopyFrom_R( pSrc->GetConstData(0, 0),
				w,ymax, pSrc->GetStride(), xmin,h+ymin);
			pBorderIm->CopyFrom_R( pSrc->GetConstData(0, h-ymin),
				w,ymin, pSrc->GetStride(), xmin,0);
			pBorderIm->CopyFrom_R( pSrc->GetConstData(0, 0),
				xmax,h, pSrc->GetStride(), w+xmin,ymin);
			pBorderIm->CopyFrom_R( pSrc->GetConstData(w-xmin,0),
				xmin,h, pSrc->GetStride(), 0,ymin);
		}
	}

	return 0;
}

int Filter::Go(const Image8u*pSrc, Image8u* pDst)
{
	int nChannels = pSrc->GetChannelCount();
	if (nChannels != pDst->GetChannelCount()) return -1;

//	pSrcData = 

	IppiSize dstSize = {
		IPP_MIN(pSrc->GetSize().width, pDst->GetSize().width),
		IPP_MIN(pSrc->GetSize().height, pDst->GetSize().height)
	};
	if (borderMode_ == NONE)
	{
		dstSize.width = 
			IPP_MIN(pSrc->GetSize().width-GetXMinBorder(), pDst->GetSize().width);
		dstSize.height = 
			IPP_MIN(pSrc->GetSize().height-GetYMinBorder(), pDst->GetSize().height);

		return Go_(pSrc->GetConstData(GetXMinBorder(),GetYMinBorder()),
			pSrc->GetStride(),
			pDst->GetData(), pDst->GetStride(),
			pDst->GetSize(), nChannels);
	}
	else 
	{
		Image8u tmp;
		MakeBorderImage(pSrc, &tmp, borderMode_);
		return Go_(tmp.GetConstData(GetXMinBorder(),GetYMinBorder()),
			tmp.GetStride(),
			pDst->GetData(), pDst->GetStride(),
			pDst->GetSize(), nChannels);
	}

//	virtual int Go_(const Ipp8u* pSrc, int srcStep,
//					   Ipp8u* pDst, int dstStep,
//					  IppiSize dstROISize, int nChannels) = 0
}
